CON
           
' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip I2C Master Interface                                     AiChip_I2cSlave_Master_002      *
' *                                                                                                     *
' *******************************************************************************************************
' *                                                                                                     *
' *     Object Name    : AiChip I2C Master Interface                                                    *
' *     Version        : 002                                                                            *
' *     Updated        : 2008-04-12                                                                     *
' *                                                                                                     *
' *     Author         : AiChip Industries                                                              *
' *                                                                                                     *
' *     Copyright      : 2008, AiChip Industries                                                        *
' *                                                                                                     *
' *     Target Chip    : Parallax Propeller Chip P8X32A ( "Propeller Mk I" )                            *
' *     Compiler Used  : Parallax Propeller Tool 1.1                                                    *
' *                                                                                                     *
' *     License        : MIT License. Please see end of this file for Terms of Use.                     *
' *                                                                                                     *
' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip is a trademark of AiChip Industries.                                                     *
' *                                                                                                     *
' *     If distributing a modified version of this software AiChip Industries requests the "AiChip"     *
' *     prefix be removed from the filename and internal text descriptions to avoid any confusion that  *
' *     may be caused by an impression that the modified software originates from AiChip Industries.    *
' *                                                                                                     *
' *******************************************************************************************************

' *******************************************************************************************************
' *                                                                                                     *
' *     HISTORY                                                                                         *
' *                                                                                                     *
' *     002     2008-04-12      Improved control line handling - No longer hard-controlled SDA          *
' *                                                                                                     *
' *     001     2008-04-11      First Public Release                                                    *
' *                                                                                                     *
' *******************************************************************************************************

PUB Version

  result := String( "AiChip_I2cSlave_Master_002" )

CON

  I2C_WR                        = 0     ' Writing to Eeprom
  I2C_RD                        = 1     ' Reading From Eeprom

  ACK                           = 0     ' Acknowledge    / More data to read
  NAK                           = 1     ' No Acknowledge / Completed data read 

  INPUT                         = 0     ' Pin Input Direction Bit
  OUTPUT                        = 1     ' Pin Output Direction Bit

  #0
  EEPROM_24LC256
  EEPROM_AT24C512

  ' 32k x 8 - Demo Board Eeprom / Spin Stamp / PropStick / PEK / DIP Plus Kit

  EEPROM_24LC256_ID             = $A0
  EEPROM_24LC256_PAGE           = 64
  EEPROM_24LC256_WRITEMS        = 5
  EEPROM_24LC256_SIZE           = 32 * 1024

  ' 64K x 8 - Proto Board Eeprom
  
  EEPROM_AT24C512_ID            = $A0
  EEPROM_AT24C512_PAGE          = 128
  EEPROM_AT24C512_WRITEMS       = 5
  EEPROM_AT24C512_SIZE          = 64 * 1024

DAT

  scl         byte              28
  sda         byte              29
  
  deviceId    byte              $A0             ' Same for both Eeprom types
  pageMask    word              64-1            ' Will work for both Eeprom types
  writeMs     byte              5               ' Same for both Eeprom Types
  shortAdr    byte              false           ' Not 8-bit address
    
VAR

  long pc

CON

' *****************************************************************************
' *                                                                           *
' *     I2C Main Routines                                                     *
' *                                                                           *
' *****************************************************************************

PRI Pause

  waitcnt( CLKFREQ / 1_000_000 * 70 + CNT )
  
PUB Open( sclPin, sdaPin, argDeviceType, bank ) ' bank = 1..N

  if --bank > 0
    bank <<= 1
  else
    bank := 0
    
  case argDeviceType
    EEPROM_24LC256  : OpenRaw( sclPin, sdaPin, bank | EEPROM_24LC256_ID,  EEPROM_24LC256_PAGE,  EEPROM_24LC256_WRITEMS  ) 
    EEPROM_AT24C512 : OpenRaw( sclPin, sdaPin, bank | EEPROM_AT24C512_ID, EEPROM_AT24C512_PAGE, EEPROM_AT24C512_WRITEMS ) 

PUB OpenRaw( sclPin, sdaPin, argDeviceId, argPageSize, argWriteMs )

  ' if PageSize < 0 then short ( 8-bit addressing )
  
  scl      := sclPin
  sda      := sdaPin
  
  deviceId := argDeviceId
  
  if argPageSize
    pageMask := ( || argPageSize ) - 1
    shortAdr := argPageSize < 0
  else
    pageMask := $FFFF_FFFF
    shortAdr := false
    
  writeMs  := argWriteMs
  
  ' Kick the I2C bus until SDA goes high

  Sda_X
  Scl_1 
  repeat 9
    Scl_0
    Scl_1
    if ina[ sda ]                     
      quit
  
  ' On exit, SDA is Input, SCL is Output High

PUB Close

   StopBit
   Scl_X
   
PUB Device( id )

  deviceId := id & $FE
  
PUB Address( adr )

  pc := adr

PUB Align( n ) ' n = 1, 2 or 4

  repeat while ( pc & (n-1) )
    pc++

' *****************************************************************************
' *                                                                           *
' *     I2C Lowest-Level Routines                                             *
' *                                                                           *
' *****************************************************************************

PRI Sda_0

  outa[ sda ] := 0
  dira[ sda ] := OUTPUT
  Pause

PRI Sda_1

  dira[ sda ] := INPUT
  Pause

PRI Sda_X

  dira[ sda ] := INPUT
  Pause
  
PRI Scl_0

  outa[ scl ] := 0
  dira[ scl ] := OUTPUT
  Pause

PRI Scl_1

  outa[ scl ] := 1
  dira[ scl ] := OUTPUT
  Pause

PRI Scl_X

  dira[ scl ] := INPUT
  Pause
  
' *****************************************************************************
' *                                                                           *
' *     I2C Low-Level Routines                                                *
' *                                                                           *
' *****************************************************************************

PUB StartBit
                   
  '       _____ ___________                       Note that we raise sda
  ' sda   _____|           |___________           then scl so that this
  '       _______________________                 can also operate as a
  ' scl   ___________|           |_____           Repeated-Start.

  Sda_1
  Scl_1
  Sda_0
  Scl_0

  
PUB StopBit

  '       _____                    _____
  ' sda   _____|__________________|
  '       ____________       ___________
  ' scl   ____________|_____|

  Sda_0
  Scl_0
  Scl_1
  Sda_1
  Sda_X
  
PUB StopBit_And_WaitForWrite

  StopBit
  if writeMs
    waitcnt( clkFreq / 1_000 * writeMs + cnt )  

PUB ReadBits( more )

  '                                                                         __________
  ' sda   _____.------------------------------------- .... ----------------|__________|___
  '               _____       _____       _____                  ______       ______
  ' scl   _______|     |_____|     |_____|     |_____ .... _____|      |_____|      |_____
  '                 7           6           5                      0           More

  Sda_X
  repeat 8
    Scl_1
    result := ( result << 1 ) | ina[ sda ]
    Scl_0
  if more
    Sda_1
  else
    Sda_0
  Scl_1
  Scl_0
  Sda_0
  
PUB WriteBits( b )

  '             __________ ___________ ___________ __      __ ____________
  ' sda   _____|__________|___________|___________|__ .... __|____________|-----------.___
  '               _____       _____       _____                  ______       ______
  ' scl   _______|     |_____|     |_____|     |_____ .... _____|      |_____|      |_____
  '                 7           6           5                      0            ACK
  repeat 8                           
    if ( b <<= 1 ) & $100
      Sda_1
    else
      Sda_0
    Scl_1
    Scl_0
  Sda_X
  Scl_1
  result := ina[ sda ]
  Scl_0
  Sda_0
  
PUB SendHeader( readOrWriteBit ) | header
  readOrWriteBit &= 1
  header := ( deviceId & $FE ) | readOrWriteBit
  StartBit
  if shortAdr
    WriteBits( header | ( ( pc >> 7 ) & $0E ) )
    if readOrWriteBit == I2C_WR
      WriteBits( pc )
  else
    WriteBits( header | ( ( pc >> 15 ) & $0E ) )
    if readOrWriteBit == I2C_WR
      WriteBits( pc >> 8 )
      WriteBits( pc )

PUB ReadSequence( numberOfBytes )
  SendHeader( I2C_WR )
  SendHeader( I2C_RD)
  case numberOfBytes
    1 : result := ReadBits( NAK )
    2 : result := ReadBits( ACK )
        result |= ReadBits( NAK ) << 8
    4 : result := ReadBits( ACK )
        result |= ReadBits( ACK ) << 8
        result |= ReadBits( ACK ) << 16
        result |= ReadBits( NAK ) << 24
  StopBit
  pc += numberOfBytes

PUB WriteSequence( n, numberOfBytes )
  SendHeader( I2C_WR )
  repeat numberOfBytes
    WriteBits( n )
    n >>= 8
  StopBit_And_WaitForWrite
  pc += numberOfBytes
    
' *****************************************************************************
' *                                                                           *
' *     I2C Reading                                                           *
' *                                                                           *
' *****************************************************************************

PUB ReadNextByteArray( arrayPtr, arraySize )
  SendHeader( I2C_WR )               
  SendHeader( I2C_RD )
  repeat arraySize-1
    byte[ arrayPtr++] := ReadBits( ACK )
  byte[ arrayPtr] := ReadBits( NAK )
  pc += arraySize
  
PUB ReadNextWordArray( arrayPtr, arraySize )

  ReadNextByteArray( arrayPtr, arraySize << 1 )

PUB ReadNextLongArray( arrayPtr, arraySize )

  ReadNextByteArray( arrayPtr, arraySize << 2 )

PUB ReadNextByte
  return ReadSequence( 1 )
  
PUB ReadNextWord

  if ( pc & 1 ) == 0
    result := ReadSequence( 2 )
  else
    result := ReadNextByte
    result |= ReadNextByte << 8 

PUB ReadNextLong

  if ( pc & 3 ) == 0
    result := ReadSequence( 4 )
  else
    result := ReadNextWord
    result |= ReadNextWord << 16

PUB ReadByte( adr )

  Address( adr )
  return ReadNextByte

PUB ReadWord( adr )

  Address( adr )
  return ReadNextWord

PUB ReadLong( adr )

  Address( adr )
  return ReadNextLong

PUB ReadByteArray( adr, arrayPtr, arraySize )

  Address( adr )
  ReadNextByteArray( arrayPtr, arraySize )  

PUB ReadWordArray( adr, arrayPtr, arraySize )

  Address( adr )
  ReadNextWordArray( arrayPtr, arraySize )

PUB ReadLongArray( adr, arrayPtr, arraySize )

  Address( adr )
  ReadNextLongArray( arrayPtr, arraySize )

' *****************************************************************************
' *                                                                           *
' *     I2C Writing                                                           *
' *                                                                           *
' *****************************************************************************

PUB WriteNextByteArray( arrayPtr, arraySize )
  SendHeader( I2C_WR )
  repeat while arraySize
    WriteBits( byte[ arrayPtr++] )
    pc++
    arraySize--
    if ( pc & pageMask ) == 0
      StopBit_And_WaitForWrite
      if arraySize
        SendHeader( I2C_WR )          
  if ( pc & pageMask )
    StopBit_And_WaitForWrite

PUB WriteNextWordArray( arrayPtr, arraySize )

  WriteNextByteArray( arrayPtr, arraySize << 1 )

PUB WriteNextLongArray( arrayPtr, arraySize )

  WriteNextByteArray( arrayPtr, arraySize << 2 )
   
PUB WriteNextByte( n ) : data
  WriteSequence(n,1)

PUB WriteNextWord( n )
  if ( pc & 1 ) == 0
    WriteSequence( n, 2 )
  else
    WriteNextByte( n )
    WriteNextByte( n >> 8 )

PUB WriteNextLong( n )
           
  if ( pc & 3 ) == 0
    WriteSequence( n, 4 )
  else
    WriteNextWord( n )
    WriteNextWord( n >> 16 )

PUB WriteByte( adr , n )

  Address( adr )
  WriteNextByte( n )

PUB WriteWord( adr , n )

  Address( adr )
  WriteNextWord( n )

PUB WriteLong( adr , n )

  Address( adr )
  WriteNextLong( n )

PUB WriteByteArray( adr, arrayPtr, arraySize )

  Address( adr )
  WriteNextByteArray( arrayPtr, arraySize )

PUB WriteWordArray( adr, arrayPtr, arraySize )

  Address( adr )
  WriteNextWordArray( arrayPtr, arraySize )

PUB WriteLongArray( adr, arrayPtr, arraySize )

  Address( adr )
  WriteNextLongArray( arrayPtr, arraySize )

CON

' *******************************************************************************************************
' *                                                                                                     *
' *     Terms of Use : MIT License                                                                      *
' *                                                                                                     *
' *******************************************************************************************************
'
' Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
' associated documentation files (the "Software"), to deal in the Software without restriction, including
' without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
' copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to
' the following conditions:                                                                   
'
' The above copyright notice and this permission notice shall be included in all copies or substantial
' portions of the Software.                                                                                             
'                                                                                                                  
' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
' LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
' NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
' WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
' SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

CON

' *******************************************************************************************************
' *                                                                                                     *
' *     End of AiChip I2C Master Interface                                                              *
' *                                                                                                     *
' *******************************************************************************************************