CON
           
' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip I2C Slave Interface                                       AiChip_I2cSlave_Slave_002      *
' *                                                                                                     *
' *******************************************************************************************************
' *                                                                                                     *
' *     Object Name    : AiChip I2C Slave Interface ( Emulates 512 byte Eeprom )                        *
' *     Version        : 002                                                                            *
' *     Updated        : 2008-04-12                                                                     *
' *                                                                                                     *
' *     Author         : AiChip Industries                                                              *
' *                                                                                                     *
' *     Copyright      : 2008, AiChip Industries                                                        *
' *                                                                                                     *
' *     Target Chip    : Parallax Propeller Chip P8X32A ( "Propeller Mk I" )                            *
' *     Compiler Used  : Parallax Propeller Tool 1.1                                                    *
' *                                                                                                     *
' *     License        : MIT License. Please see end of this file for Terms of Use.                     *
' *                                                                                                     *
' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip is a trademark of AiChip Industries.                                                     *
' *                                                                                                     *
' *     If distributing a modified version of this software AiChip Industries requests the "AiChip"     *
' *     prefix be removed from the filename and internal text descriptions to avoid any confusion that  *
' *     may be caused by an impression that the modified software originates from AiChip Industries.    *
' *                                                                                                     *
' *******************************************************************************************************

' *******************************************************************************************************
' *                                                                                                     *
' *     HISTORY                                                                                         *
' *                                                                                                     *
' *     002     2008-04-12      PASM version of I2C Slave included - Not working yet                    *
' *                                                                                                     *
' *     001     2008-04-11      First Public Release                                                    *
' *                                                                                                     *
' *******************************************************************************************************

PUB Version

  result := String( "AiChip_I2cSlave_Slave_002" )

CON
      
' .-----------------------------------------------------------------------------------------------------.
' |     Object Handlers                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

CON

  STACK_SIZE = 40
  
VAR

  byte CogNumber

  long stack[ STACK_SIZE ]

PUB StartSpin( sclPin, sdaPin, id, a2, a1, a0 ) | i

  Stop

  id := ( id & $FE ) | a2 << 3 | a1 << 2 | a0 << 1

  repeat i from 0 to STACK_SIZE-1
    stack[ i ] := $D0D0_F00D
    
  cogNumber := CogNew( I2cSlave( sclPin, sdaPin, id ), @stack ) + 1

PUB StartPasm( sclPin, sdaPin, id, a2, a1, a0 )

  Stop

  param_sclPin   := sclPin
  param_sdaPin   := sdaPin
  param_DeviceId := ( id & $FE ) | a2 << 3 | a1 << 2 | a0 << 1

  cogNumber := CogNew( @Pasm_I2cSlave, @param ) + 1

PUB Stop

  if cogNumber
    CogStop( cogNumber~ - 1 )

PUB StackUsed

  repeat while stack[ result++ ] <> $D0D0_F00D

CON
      
' .-----------------------------------------------------------------------------------------------------.
' |     Actual I2C Slave Object                                                                         |
' `-----------------------------------------------------------------------------------------------------'

CON

  ADDRESS_MASK = $1FF

VAR
  
  byte bitInputs
  byte bitStream
  
  byte scl
  byte sda
  byte sclShift
  byte sdaShift

  byte deviceId

  byte ram[ ADDRESS_MASK+1 ]

CON
      
' .-----------------------------------------------------------------------------------------------------.
' |     Main I2C Slave Handling Loop                                                                    |
' `-----------------------------------------------------------------------------------------------------'
    
PRI I2cSlave( sclPin, sdaPin, id ) | inputs

  scl := sclPin
  sda := sdaPin

  sclShift := scl-1
  sdaShift := sda

  deviceId := id & $FE
  
  inputs    := INA
  bitInputs := ( ( inputs >> sclShift ) & 2 ) | ( ( inputs >> sdaShift ) & 1 )
  bitStream := ( bitInputs << 4 ) | ( bitInputs << 2 ) | bitInputs
  
  \Idle
  repeat
    \GotStartBit

CON
      
' .-----------------------------------------------------------------------------------------------------.
' |     I2C Command Handling                                                                            |
' `-----------------------------------------------------------------------------------------------------'

PRI GotStartBit | id, adr, num

  ' We only get here when a start bit has been detected
  
  if ( ( id := GetByte ) & $FE ) == deviceId

    ' Command directed to us
    
    if id & 1

      ' Read command
      
      Ack
      repeat
        SendByte( ram[ adr++ & ADDRESS_MASK ] )
      until GetBit

    else

      ' Write Command
      
      Ack
      adr := GetByte << 8
      Ack
      adr |= GetByte
      Ack
      repeat
        num := GetByte
        ram[ adr++ & ADDRESS_MASK ] := num
        Ack

  ' Go into Idle mode - Wait for the next start bit
  
  Idle

PRI Idle

  repeat
    GetBit

CON
      
' .-----------------------------------------------------------------------------------------------------.
' |     Output Routines                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

PRI Ack

  SendBit( 1 )

PRI Nak

  SendBit( 0 )
    
PRI SendByte( n )

  repeat 8
    SendBit( n >> 7 )
    n <<= 1
    
PRI SendBit( n )

  repeat while INA[ scl ]
  DIRA[ sda ] := ( n & 1 ) ^ 1
  repeat until INA[ scl ]
  repeat while INA[ scl ]
  DIRA[ sda ] := 0

CON
      
' .-----------------------------------------------------------------------------------------------------.
' |     Input Routines                                                                                  |
' `-----------------------------------------------------------------------------------------------------'
  
PRI GetByte

  repeat 8
    result := ( result << 1 ) | GetBit
     
PRI GetBit

  ' Check the bit stream to see if we can detect a sequence which indicates something we can understand,
  '
  ' A start bit will abort which will fall rigt back to the outer loop which will cause the GitStartBit
  ' routine to be executed.
  '
  ' If a stop bit is detected we stop doing anything until we get a subsequent start bit, and handle it
  ' as a start bit, as per above.
  '
  ' Note that the bit stream must have 'clean and identifiable transitions', ie only one line must change
  ' at a time, for example, SDA must be set before SCLK is raised, SCLK must be lowered before SDA is
  ' changed to detect a valid data bit.
  
  repeat
    case GetBitStream
      'CD CD CD
      %11_10_00 : Abort                                                         ' Start Bit
      %00_10_11 : repeat until GetBitStream == %11_10_00                        ' Stop bit
                  Abort                                                         
      %00_10_00 : return 0                                                      ' Bit 0
      %01_11_01 : return 1                                                      ' Bit 1
  
PRI GetBitStream | inputs, thisBits

  ' Build up an incomming bit stream which will have three samples in it of the last three different
  ' states. The bit stream is in the form %CD_CD_CD where C is the SCLK state, D is the SDA state. The
  ' earliest state is on the left, the current state is on the right.
  
  repeat
    inputs   := INA
    thisBits := ( ( inputs >> sclShift ) & 2 ) | ( ( inputs >> sdaShift ) & 1 ) 
  until thisBits <> bitInputs
  
  bitInputs := thisBits

  result := bitStream := ( ( bitStream << 2 ) & %11_11_00 ) | bitInputs

CON

' *******************************************************************************************************
' *                                                                                                     *
' *     PASM Handler                                                                                    *
' *                                                                                                     *
' *******************************************************************************************************

VAR                     long    param[0]
  
                        long    param_sclPin
                        long    param_sdaPin
                        long    param_deviceId
  
DAT                     org     $000

'                       .-------------------------------------------------------------------------------.
'                       |       initialise the parameters                                               |
'                       `-------------------------------------------------------------------------------'

Pasm_I2cSlave           mov     pasm_idx,PAR

                        call    #Pasm_GetParam                                  
                        mov     pasm_sclMask,#1
                        shl     pasm_sclMask,pasm_byte                        

                        call    #Pasm_GetParam                                  
                        mov     pasm_sdaMask,#1
                        shl     pasm_sdaMask,pasm_byte                        

                        call    #Pasm_GetParam                                  
                        and     pasm_byte,#$FE
                        mov     pasm_deviceWr,pasm_byte
                        or      pasm_byte,#1
                        mov     pasm_deviceRd,pasm_byte

'                       .-------------------------------------------------------------------------------.
'                       |       Initialise the incoming bit stream                                      |
'                       `-------------------------------------------------------------------------------'

                        mov     pasm_inputs,INA
                        test    pasm_inputs,pasm_sclMask WC
                        muxc    pasm_thisBits,#2
                        test    pasm_inputs,pasm_sdaMask WC
                        muxc    pasm_thisBits,#1
                        and     pasm_thisBits,#%11

                        mov     pasm_bitInputs,pasm_thisBits
                        
                        mov     pasm_bitStream,pasm_bitInputs
                        shl     pasm_bitStream,#2
                        or      pasm_bitStream,pasm_bitInputs

'                       .-------------------------------------------------------------------------------.
'                       |       Wait for a start bit                                                    |
'                       `-------------------------------------------------------------------------------'
                        
Pasm_Idle               call    #Pasm_GetBit
                        jmp     #Pasm_Idle

'                       .-------------------------------------------------------------------------------.
'                       |       Determine if command is for us or not                                   |
'                       `-------------------------------------------------------------------------------'

Pasm_GotStartBit        call    #Pasm_GetByte
                        cmp     pasm_byte,pasm_deviceRd WZ
                  IF_Z  jmp     #Pasm_Read
                        cmp     pasm_byte,pasm_deviceWr WZ
                  IF_Z  jmp     #Pasm_Write
                        jmp     #Pasm_Idle

'                       .-------------------------------------------------------------------------------.
'                       |       Handle a Read command                                                   |
'                       `-------------------------------------------------------------------------------'

Pasm_Read               call    #Pasm_Ack

:Loop                   call    #Pasm_LoadByte
                        call    #Pasm_SendByte
                        call    #Pasm_GetBit
                        tjnz    pasm_bit,#:Loop
                        jmp     #Pasm_Idle

'                       .-------------------------------------------------------------------------------.
'                       |       Handle a Write command                                                  |
'                       `-------------------------------------------------------------------------------'

Pasm_Write              call    #Pasm_Ack

                        call    #Pasm_GetByte
                        mov     pasm_adr,pasm_byte
                        call    #Pasm_Ack
                        
                        call    #Pasm_GetByte
                        shl     pasm_adr,#8
                        or      pasm_adr,pasm_byte
                        call    #Pasm_Ack
                        
:Loop                   call    #Pasm_GetByte
                        Call    #Pasm_SaveByte
                        call    #Pasm_Ack                                              
                        jmp     #:Loop

'                       .-------------------------------------------------------------------------------.
'                       |       Handle saving and loading from emulated Eeprom                          |
'                       `-------------------------------------------------------------------------------'

Pasm_SaveByte           mov     pasm_idx,pasm_adr
                        and     pasm_idx,#$0FF
                        add     pasm_idx,#pasm_ram     
                        movd    :Dst,pasm_idx
                        add     pasm_adr,#1
:Dst                    mov     0-0,pasm_byte
Pasm_SaveByte_Ret       ret

Pasm_LoadByte           mov     pasm_idx,pasm_adr
                        and     pasm_idx,#$0FF
                        add     pasm_idx,#pasm_ram     
                        movs    :Src,pasm_idx
                        add     pasm_adr,#1
:Src                    mov     pasm_byte,0-0
Pasm_LoadByte_Ret       ret

'                       .-------------------------------------------------------------------------------.
'                       |       Send Ack or Nak                                                         |
'                       `-------------------------------------------------------------------------------'
                                                
Pasm_Ack                mov     pasm_byte,#$80
                        call    #Pasm_SendBit
Pasm_Ack_Ret            ret

Pasm_Nak                mov     pasm_byte,#0
                        call    #Pasm_SendBit
Pasm_Nak_Ret            ret

'                       .-------------------------------------------------------------------------------.
'                       |       Send a byte                                                             |
'                       `-------------------------------------------------------------------------------'

Pasm_SendByte           mov     pasm_bitCount,#8

:Loop                   call    #Pasm_SendBit
                        shl     pasm_byte,#1
                        djnz    pasm_bitCount,#:Loop

Pasm_SendByte_Ret       ret

'                       .-------------------------------------------------------------------------------.
'                       |       Send a bit                                                              |
'                       `-------------------------------------------------------------------------------'

Pasm_SendBit            test    pasm_sclMask,INA WC
                  IF_C  jmp     #$-1

                        test    pasm_byte,#$80 WC
                  IF_NC mov     DIRA,pasm_sdaMask
                  
                        test    pasm_sclMask,INA WC
                  IF_NC jmp     #$-1
                  
                        test    pasm_sclMask,INA WC
                  IF_C  jmp     #$-1
                  
                        mov     DIRA,#0

Pasm_SendBit_Ret        ret

'                       .-------------------------------------------------------------------------------.
'                       |       Read a byte                                                             |
'                       `-------------------------------------------------------------------------------'

Pasm_GetByte            mov     pasm_bitCount,#8
                        mov     pasm_byte,#0
                        
:Loop                   call    #Pasm_GetBit
                        shl     pasm_byte,#1
                        or      pasm_byte,pasm_bit
                        
                        djnz    pasm_bitCount,#:Loop

Pasm_GetByte_Ret        ret

'                       .-------------------------------------------------------------------------------.
'                       |       Determine bit or event from the incoming bit stream                     |
'                       `-------------------------------------------------------------------------------'

Pasm_GetBit             call    #Pasm_GetBitStream

                        cmp     pasm_bitStream,#%11_10_00 WZ
                  IF_Z  jmp     #Pasm_GotStartBit

                        cmp     pasm_bitStream,#%00_10_11 WZ
                  IF_NZ jmp     #:Check0

:WaitForStart           call    #Pasm_GetBitStream
                        cmp     pasm_bitStream,#%11_10_00 WZ
                  IF_Z  jmp     #Pasm_GotStartBit
                        jmp     #:WaitForStart

:Check0                 cmp     pasm_bitStream,#%00_10_00 WZ
                  IF_NZ jmp     #:Check1

                        mov     pasm_bit,#0
                        jmp     #Pasm_GetBit_Ret
                        
:Check1                 cmp     pasm_bitStream,#%01_11_01 WZ
                  IF_NZ jmp     #Pasm_GetBit

                        mov     pasm_bit,#1
                        
Pasm_GetBit_Ret         ret

'                       .-------------------------------------------------------------------------------.
'                       |       Read a bit pair into the bit stream                                     |
'                       `-------------------------------------------------------------------------------'

Pasm_GetBitStream       mov     pasm_inputs,INA
                        test    pasm_inputs,pasm_sclMask WC
                        muxc    pasm_thisBits,#2
                        test    pasm_inputs,pasm_sdaMask WC
                        muxc    pasm_thisBits,#1
                       
                        cmp     pasm_thisBits,pasm_bitInputs WZ
                  IF_Z  jmp     #Pasm_GetBitStream

                        mov     pasm_bitInputs,pasm_thisBits

                        shl     pasm_bitStream,#2
                        and     pasm_bitStream,#%11_11_00
                        or      pasm_bitStream,pasm_bitInputs
                        
Pasm_GetBitStream_Ret   ret

'                       .-------------------------------------------------------------------------------.
'                       |       Get a parameter                                                         |
'                       `-------------------------------------------------------------------------------'

Pasm_GetParam           rdlong  pasm_byte,pasm_idx
                        add     pasm_idx,#4
Pasm_GetParam_Ret       ret

'                       .-------------------------------------------------------------------------------.
'                       |       Variables                                                               |
'                       `-------------------------------------------------------------------------------'

pasm_sclMask            res     1
pasm_sdaMask            res     1
pasm_deviceRd           res     1
pasm_deviceWr           res     1

pasm_inputs             res     1
pasm_thisBits           res     1
pasm_bitInputs          res     1
pasm_bitStream          res     1

pasm_bitCount           res     1
pasm_bit                res     1
pasm_byte               res     1

pasm_adr                res     1
pasm_idx                res     1

pasm_ram                res     256

                        fit     $1F0

CON

' *******************************************************************************************************
' *                                                                                                     *
' *     Terms of Use : MIT License                                                                      *
' *                                                                                                     *
' *******************************************************************************************************
'
' Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
' associated documentation files (the "Software"), to deal in the Software without restriction, including
' without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
' copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to
' the following conditions:                                                                   
'
' The above copyright notice and this permission notice shall be included in all copies or substantial
' portions of the Software.                                                                                             
'                                                                                                                  
' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
' LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
' NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
' WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
' SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

CON

' *******************************************************************************************************
' *                                                                                                     *
' *     End of AiChip I2C Slave Interface                                                               *
' *                                                                                                     *
' *******************************************************************************************************   